function [label, Tree] = ntp(G,k)
% The normalized tree partitioning algorithm in the CVPR'08 paper:
%   Wang, J. and Jia, Y. and Hua, X.S. and Zhang, C. and Quan, L., 
%   Normalized tree partitioning for image segmentation. CVPR 2008.
% Input: G - the sparse graph structure, where G(i,j) is the weight between
%        the vertex i and vertex j if they are connected. If i and j are
%        not connected, G(i,j) = 0. We assume that the weights are
%        non-negative.
%        k - the number of partitionings needed to be computed. In default
%        it is set to be 2.
% Output: label - the label for each vertex in the graph.
%
% Note that the code is slow since Matlab is not good at processing loops,
% so do not use this to compare speed. Also, in this code we simply used
% the minimum spanning tree instead of the minimum entropy spanning tree,
% which might be slightly different from the paper.
%
% Yangqing Jia, jiayq@eecs.berkeley.edu, 2010.

if nargin==1
    k = 2;
end

% sanity check
if k < 2
    error('ntp: the number of partitionings should be at least 2.');
end

% first, compute the maximum spanning tree. Since conventionally we use the
% minimum spanning tree idea, we will convert the problem to a minimum
% spanning tree problem.
Tree = mst(-G);
Tree = -Tree;

if k==2
    label = ntp_base(Tree);
else
    % recursively do the partitioning until we have reached the number of
    % partitions defined by k.
    % initialize the labels
    label = ones(1,size(G,1));
    for iter = 1:k-1
        % for each iter, we need to calculate the partitioning score for each
        % subtree, and pick the best one
        ntpVals = zeros(1,iter);
        for idx = 1:iter
            subIdx = find(label==idx);
            subTree = Tree(subIdx,subIdx);
            [ignore,ntpVals(idx)] = ntp_base(subTree);
        end
        [ignore,bestIdx] = min(ntpVals);
        % a little waste of time recomputing the cut, but let's keep it this
        % way
        subIdx = find(label==bestIdx);
        subTree = Tree(subIdx,subIdx);
        [subLabel,ignore] = ntp_base(subTree);
        label(subIdx(subLabel==2)) = iter+1;
    end
end

end % end of function ntp