function [label,score] = ntp_base(Tree)
% the base function of ntp that partitions a tree to two partitions using
% the normalized tree partitioning algorithm, via dynamic programming.
% This function is called by the ntp function.
% 
% Yangqing Jia, jiayq@eecs.berkeley.edu, 2010.

% This function did a little hack by using the java class LinkedList as the
% queue implementation of the queue.

import java.util.LinkedList;

parents = getParents(Tree);
leaves = getLeaves(Tree);
root = getRoot(Tree);
% borrow the java LinkedList code to carry out the queue operation
queue = LinkedList();
for leaf = 1:length(leaves)
    queue.offer(leaves(leaf));
end
% compute the association values
assoc = zeros(1,size(Tree,1));
Nc = getNchildren(Tree);
Nc(Nc==0) = -1; % -1 denotes that the vertex has been processed
while ~isempty(queue.peek())
    current = queue.poll();
    if current==root
        continue;
    end
    parent = parents(current);
    Nc(parent) = Nc(parent) - 1;
    if Nc(parent) == 0
        Nc(parent) = -1;
        queue.offer(parent);
    end
    assoc(parent) = assoc(parent) + assoc(current) + 2*Tree(current,parent);
end
clear queue;
% compute the normalized cuts value
assocRoot = assoc(root);
ncut = zeros(1,size(Tree,1));
ncut(root) = Inf;
for idx = 1:size(Tree,1)
    if idx==root
        continue;
    end
    % compute the ncut value for vertex i
    aIdx = assoc(idx);
    aR_idx = assocRoot - aIdx - 2*Tree(idx,parents(idx));
    cut = Tree(idx,parents(idx));
    ncut(idx) = cut/(aIdx+cut) + cut/(aR_idx+cut);
end
[score,bestIdx] = min(ncut);

% output the labels. to avoid stack overflow we will again use the queue
% instead of using recursive functions.
label = ones(1,size(Tree,1));
queue = LinkedList();
queue.offer(bestIdx);
while ~isempty(queue.peek())
    current = queue.poll();
    label(current) = 2;
    children = getChildren(Tree,current);
    for child = 1:length(children)
        queue.offer(children(child));
    end
end

end % end of function ntp_base.

%% %%%%%%%%%%%%%%%%%%%%%%%
% sub function: get the leaves from the tree
function [leaves] = getLeaves(Tree)
leaves = find(sum(Tree)==0);
end

% sub function: get the root from the tree
function [root] = getRoot(Tree)
root = find(sum(Tree,2)==0);
end

% sub function: get the parent for each vertex in the tree
% The parent of the root vertex would be zero.
function [parents] = getParents(Tree)
parents = zeros(1,size(Tree,1));
[i,j,ignore] = find(Tree);
parents(i) = j;
end

% sub function: get the children for the vertex idx
function [children] = getChildren(Tree,idx)
children = find(Tree(:,idx)~=0);
end

% sub function: get the number of children for the vertices
function [Nc] = getNchildren(Tree)
Nc = sum(Tree~=0);
end


